from .config import pg, username
from .posterino import submitRound
from .hosterino import checkAnswers
from .Utils.utils import waitForApproval, approved, postDelay, randomColor
from .Loader.loader import getRound
from sty import fg
from importlib.metadata import version
from update_checker import UpdateChecker
import time

v = version('botterino')
checker = UpdateChecker()
result = checker.check('botterino', v)
if result:
    print(f'{fg.yellow}{result}')
    print(f'{fg.yellow}run pip install --upgrade botterino to update')

def checkType(r):
    if 'tolerance' in r and 'answer' in r:
        return "automatic"
    if 'tolerances' in r and 'answers' in r:
        return "automatic"
    if 'manual' in r:
        return 'x wrong guesses, manual correct'
    return 'manual'

while True:
    print(f'{fg.yellow}Waiting for {username} to win a round... 🐌')
    waitForApproval()
    print(f'{fg.blue}Congrats on a well deserved win {username}! ⭐')
    r = getRound()
    while not r:
        print(f'{fg.red}No rounds in round file! checking again in 10s')
        time.sleep(10)
        r = getRound()
    submission = submitRound(r)
    print(f'{randomColor()}Your round was posted to https://reddit.com{submission.permalink}')
    print(f'{fg.magenta}Round \'{r["title"]}\' posted in {postDelay()}s')
    print(f'{fg.cyan}Checking Answers: {checkType(r)}...')
    checkAnswers(r, submission)
    while approved():
        continue
    after = r.get('after')
    if after:
        submission.reply(after)
        print(f'{randomColor()}Posted your message after the round: {after}')



import praw
import os
from . import botfiles
from sty import fg

debug = False
roundfile = botfiles.rounds
archivefile = botfiles.archive
cwd = os.getcwd()
try:
    os.chdir(botfiles.botconfig)
    reddit = praw.Reddit('botterino')
    print(f'{fg.green}Successfully logged into reddit as {reddit.user.me()}')
except Exception as e:
    print(f'{fg.red}Unable to login to reddit. Please check {botfiles.prawconfig}')
    print(f'{fg.red}{e}')
finally:
    os.chdir(cwd)

pg = reddit.subreddit('itoxtestingfacility' if debug else 'picturegame')

username = str(reddit.user.me())

donotreply = {
    'achievements-bot',
    username.lower(),
    'r-picturegame',
    'imreallycuriousbird',
}

incorrect = '❌'



import itertools

from .Retry import retry

POLL_LIMIT = 50

class FifoSet:
    def __init__(self, size):
        self.size = size
        self._fifo = []
        self._set = set()

    def __contains__(self, item):
        return item in self._set

    def add(self, item):
        if len(self._set) == self.size:
            self._set.remove(self._fifo.pop(0))
        self._fifo.append(item)
        self._set.add(item)

class CommentWrapper:
    def __init__(self, func1, func2):
        self.f1 = func1
        self.f2 = func2
    def __call__(self, *args, **kwargs):
        return itertools.chain(self.f1(*args, **kwargs), self.f2(*args, **kwargs))


class RedditPoller:
    def __init__(self, function, before = None):
        self.function = function
        self.seenNames = FifoSet(POLL_LIMIT * 1000)
        self.beforeName = before

    def getLatest(self):
        while True:
            newestName = None
            for item in self._poll():
                if item.name in self.seenNames:
                    continue
                self.seenNames.add(item.name)
                newestName = item.name
                yield item

            self.beforeName = newestName
            yield None

    @retry
    def _poll(self):
        return reversed(list(self.function(limit = POLL_LIMIT, params = { "before": self.beforeName })))


from pathlib import Path
from ..config import roundfile, archivefile
import ruamel.yaml

yaml = ruamel.yaml.YAML()
yaml.preserve_quotes = True
yaml.allow_duplicate_keys = True

rounds = Path(roundfile)
archive = Path(archivefile)

def getRound():
    x = yaml.load(rounds)
    if not x:
        return None
    k = next(iter(x))
    top = x.pop(k)
    if x:
        yaml.dump(x, rounds)
    else:
        open(rounds, 'w').close()
    y = yaml.load(archive) or {}
    y[k] = top
    yaml.dump(y, archive)
    return top


import os
from pathlib import Path

class files():
    def __init__(self):
        self.home = os.path.expanduser('~')
        self.botconfig = os.path.join(self.home, 'botterino-config')
        self.roundsdir = os.path.join(self.botconfig, 'rounds')
        self.rounds = os.path.join(self.roundsdir, 'rounds.yaml')
        self.archive = os.path.join(self.roundsdir, 'archive.yaml')
        self.prawconfig = os.path.join(self.botconfig, 'praw.ini')

botfiles = files()
dirs = [botfiles.botconfig, botfiles.roundsdir]
files = [botfiles.rounds, botfiles.archive, botfiles.prawconfig]
for d in dirs:
    Path(d).mkdir(parents=True, exist_ok=True)
for f in files:
    with open(f, 'a+'):
        pass



from .hosterino import checkAnswers
from .config import pg
from .Utils.utils import randomColor
from .Loader import loader
from sty import fg
import time


r = loader.getRound()
while not r:
    print(f'{fg.red}No rounds in round file! checking again in 10s')
    time.sleep(10)
    r = loader.getRound()

submission = next(iter(pg.new()))
print(f'{randomColor()}Checking answers on https://reddit.com{submission.permalink}')
checkAnswers(r, submission)
after = r.get('after')
if after:
    submission.reply(after)


from geopy.distance import distance
from geopy.point import Point
from .config import donotreply, incorrect, reddit, username, pg
from itertools import permutations
import re
from sty import fg
from .Utils.utils import decimal, getComments, getDistance, randomColor, randomColorWithAuthor


def withinTolerance(guess, answer, tolerance):
    return distance(guess, answer).m <= tolerance

def checkMultipleCoordinates(guess, answers, tolerances):
    guesser = guess.author.name
    answers = [Point(a) for a in answers]
    match = re.findall(decimal, guess.body)
    if len(match) != len(answers):
        # TODO print a message here
        print(f'{randomColorWithAuthor(guesser)}{guesser}\'s guess {guess.body} was incorrect')
        return False
    try:
        points = [Point(f'{lat},{lon}') for lat, lon in match]
        points = permutations(points)
    except Exception as e:
        print(f'{randomColor()}Something happened: ', e, 'it probably does not matter')
        return False

    results = [[withinTolerance(p, a, t) for p, a, t in zip(ps, answers, tolerances)] for ps in points]
    results = [all(r) for r in results]
    result = any(results)
    if not result:
        print(f'{randomColorWithAuthor(guesser)}{guesser}\'s guess {guess.body} was incorrect')
    return result

def checkCoordinates(guess, answer, tolerance):
    guesser = guess.author.name
    answer = Point(answer)
    error = getDistance(guess.body, answer)
    if error is None:
        print(f"{randomColorWithAuthor(guesser)}Could not find a coordinate in guess '{guess.body}' by {guesser}")
        return 'ignore'
    error = round(error, 2)
    print(f'{randomColorWithAuthor(guesser)}{guesser}\'s guess {guess.body} was {error} meters off')
    return error <= tolerance

def checkAnswers(r, submission):
    tolerance, manual, after, text, answer, tolerances, answers = float(r.get('tolerance', 0)), r.get(
        'manual'), r.get('after'), r.get('text'), r.get('answer'), r.get('tolerances'), r.get('answers')
    if not tolerance and not tolerances:
        return
    for c in getComments(submission):
        result = True
        if tolerance:
            r = checkCoordinates(c, answer, tolerance)
            if r == 'ignore':
                continue
            result = result and r
        elif tolerances:
            tolerances = [float(t) for t in r['tolerances']]
            if len(answers) != len(tolerances):
                print('{fg.red}Refusing to check answers, number of tolerances must equal number of answers.')
            result = result and checkMultipleCoordinates(c, answers, tolerances)

        if not result:
            c.reply(incorrect)
        if result:
            if manual:
                print(f"{randomColor()}Guess '{c.body}' looks correct, but you will have to check it out.")
            else:
                plusCorrect = c.reply('+correct')
                guesser = c.author.name
                print(
                    f'{randomColorWithAuthor(guesser)}Corrected {guesser} in {plusCorrect.created_utc - c.created_utc}s')
                break
